@@ -88,8 +88,9 @@ static void print_help(void)
 	print_help_line("      --gost94-cryptopro ", digest_format, _("GOST R 34.11-94 CryptoPro"));
 	print_help_line("      --ripemd160  ", digest_format, "RIPEMD-160");
 	print_help_line("      --has160     ", digest_format, "HAS-160");
-	print_help_line("      --edonr256, --edonr512  ", digest_format, "EDON-R 256/512");
-	print_help_line("      --snefru128, --snefru256  ", digest_format, "SNEFRU-128/256");
+	print_help_line("      --blake2s,   --blake2b   ", digest_format, "BLAKE2S/BLAKE2B");
+	print_help_line("      --edonr256,  --edonr512  ", digest_format, "EDON-R 256/512");
+	print_help_line("      --snefru128, --snefru256 ", digest_format, "SNEFRU-128/256");
 	print_help_line("  -a, --all        ", _("Calculate all supported hash functions.\n"));
 	print_help_line("  -c, --check      ", _("Check hash files specified by command line.\n"));
 	print_help_line("  -u, --update=<file> ", _("Update the specified hash file.\n"));
@@ -135,7 +136,7 @@ static void print_help(void)
  */
 static void list_hashes(void)
 {
-	int id;
+	unsigned id;
 	for (id = 1; id < RHASH_ALL_HASHES; id <<= 1) {
 		const char* hash_name = rhash_get_name(id);
 		if (hash_name) rsh_fprintf(rhash_data.out, "%s\n", hash_name);
@@ -395,6 +396,8 @@ cmdline_opt_t cmdline_opt[] =
 	{ F_UFLG,   0,   0, "snefru256", &opt.sum_flags, RHASH_SNEFRU256 },
 	{ F_UFLG,   0,   0, "edonr256",  &opt.sum_flags, RHASH_EDONR256 },
 	{ F_UFLG,   0,   0, "edonr512",  &opt.sum_flags, RHASH_EDONR512 },
+	{ F_UFLG,   0,   0, "blake2s",   &opt.sum_flags, RHASH_BLAKE2S },
+	{ F_UFLG,   0,   0, "blake2b",   &opt.sum_flags, RHASH_BLAKE2B },
 	{ F_UFLG, 'L',   0, "ed2k-link", &opt.sum_flags, OPT_ED2K_LINK },
 
 	/* output formats */
@@ -434,6 +437,7 @@ cmdline_opt_t cmdline_opt[] =
 	{ F_UFNC,   0,   0, "bt-announce", bt_announce, 0 },
 	{ F_TSTR,   0,   0, "bt-batch", &opt.bt_batch_file, 0 },
 	{ F_UFLG,   0,   0, "benchmark-raw", &opt.flags, OPT_BENCH_RAW },
+	{ F_UFLG,   0,   0, "no-detect-by-ext", &opt.flags, OPT_NO_DETECT_BY_EXT },
 	{ F_UFLG,   0,   0, "hex", &opt.flags, OPT_HEX },
 	{ F_UFLG,   0,   0, "base32", &opt.flags, OPT_BASE32 },
 	{ F_UFLG, 'b',   0, "base64", &opt.flags, OPT_BASE64 },
@@ -549,89 +553,125 @@ static void apply_option(options_t* opts, parsed_option_t* option)
 	}
 }
 
+#ifdef _WIN32
+# define rsh_tgetenv(name) _wgetenv(name)
+#else
+# define rsh_tgetenv(name) getenv(name)
+#endif
+#define COUNTOF(array) (sizeof(array) / sizeof(*array))
+enum ConfigLookupFlags
+{
+	ConfFlagNeedSplit = 8,
+	ConfFlagNoVars = 16
+};
+
+/**
+ * Check if a config file, specified by path subparts, is a regular file.
+ * On success the resulting path is stored as rhash_data.config_file.
+ *
+ * @param path_parts subparts of the path
+ * @param flags check flags
+ * @return 1 if the file is regular, 0 otherwise
+ */
+static int try_config(ctpath_t path_parts[], unsigned flags)
+{
+	const size_t parts_count = flags & 3;
+	tpath_t allocated = NULL;
+	ctpath_t path = NULL;
+	size_t i;
+	for (i = 0; i < parts_count; i++) {
+		ctpath_t sub_path = path_parts[i];
+		if (sub_path[0] == RSH_T('$') && !(flags & ConfFlagNoVars)) {
+			sub_path = rsh_tgetenv(sub_path + 1);
+			if (!sub_path || !sub_path[0]) {
+				free(allocated);
+				return 0;
+			}
+#ifndef _WIN32
+			/* check if the variable should be splitted */
+			if (flags == (2 | ConfFlagNeedSplit) && i == 0) {
+				tpath_t next;
+				ctpath_t parts[2];
+				parts[1] = path_parts[1];
+				sub_path = allocated = strdup(sub_path);
+				do {
+					next = strchr(sub_path, ':');
+					if (next)
+						*(next++) = '\0';
+					if (sub_path[0]) {
+						parts[0] = sub_path;
+						if (try_config(parts, COUNTOF(parts) | ConfFlagNoVars)) {
+							free(allocated);
+							return 1;
+						}
+					}
+					sub_path = next;
+				} while(sub_path);
+				free(allocated);
+				return 0;
+			}
+#endif
+		}
+		if (path) {
+			tpath_t old_allocated = allocated;
+			path = allocated = make_tpath(path, sub_path);
+			free(old_allocated);
+		} else {
+			path = sub_path;
+		}
+	}
+	assert(!rhash_data.config_file.real_path);
+	{
+		unsigned init_flags = FileInitRunFstat | (!allocated ? FileInitReusePath : 0);
+		int res = file_init(&rhash_data.config_file, path, init_flags);
+		free(allocated);
+		if (res == 0 && FILE_ISREG(&rhash_data.config_file))
+			return 1;
+		file_cleanup(&rhash_data.config_file);
+		return 0;
+	}
+}
+
 /**
  * Search for config file.
  *
- * @return the relative path to config file
+ * @return 1 if config file is found, 0 otherwise
  */
-static const char* find_conf_file(void)
+static int find_conf_file(void)
 {
 #ifndef SYSCONFDIR
 # define SYSCONFDIR "/etc"
 #endif
-#define CONFIG_FILENAME "rhashrc"
 
-	char* dir1;
-	char* path;
+#ifndef _WIN32
+	/* Linux/Unix part */
+	static ctpath_t xdg_conf_home[2] = { "$XDG_CONFIG_HOME", "rhash/rhashrc" };
+	static ctpath_t xdg_conf_default[2] = { "$HOME", ".config/rhash/rhashrc" };
+	static ctpath_t xdg_conf_dirs[2] = { "$XDG_CONFIG_DIRS", "rhash/rhashrc" };
+	static ctpath_t home_conf[2] = { "$HOME", ".rhashrc" };
+	static ctpath_t sysconf_dir[1] = { SYSCONFDIR "/rhashrc" };
 
-#ifndef _WIN32 /* Linux/Unix part */
-	/* first check for $XDG_CONFIG_HOME/rhash/rhashrc file */
-	if ( (dir1 = getenv("XDG_CONFIG_HOME")) ) {
-		dir1 = make_path(dir1, "rhash", 0);
-		path = make_path(dir1, CONFIG_FILENAME, 0);
-		free(dir1);
-		if (is_regular_file(path)) {
-			rsh_vector_add_ptr(opt.mem, path);
-			return (conf_opt.config_file = path);
-		}
-		free(path);
-	}
-	/* then check for $HOME/.rhashrc file */
-	if ( (dir1 = getenv("HOME")) ) {
-		path = make_path(dir1, ".rhashrc", 0);
-		if (is_regular_file(path)) {
-			rsh_vector_add_ptr(opt.mem, path);
-			return (conf_opt.config_file = path);
-		}
-		free(path);
-	}
-	/* then check for global config */
-	path = SYSCONFDIR "/" CONFIG_FILENAME;
-	if (is_regular_file(path)) {
-		return (conf_opt.config_file = path);
-	}
+	return (try_config(xdg_conf_home, COUNTOF(xdg_conf_home)) ||
+			try_config(xdg_conf_default, COUNTOF(xdg_conf_default)) ||
+			try_config(xdg_conf_dirs, COUNTOF(xdg_conf_dirs) | ConfFlagNeedSplit) ||
+			try_config(home_conf, COUNTOF(home_conf)) ||
+			try_config(sysconf_dir, COUNTOF(sysconf_dir)));
 
 #else /* _WIN32 */
-	wchar_t* program_dir = get_program_dir();
 
-	/* first check for the %APPDATA%\RHash\rhashrc config */
-	if ( (dir1 = getenv("APPDATA")) ) {
-		dir1 = make_path(dir1, "RHash", 0);
-		path = make_path(dir1, CONFIG_FILENAME, 0);
-		free(dir1);
-		if (is_regular_file(path)) {
-			rsh_vector_add_ptr(opt.mem, path);
-			return (conf_opt.config_file = path);
-		}
-		free(path);
+	static ctpath_t app_data[2] = { L"$APPDATA", L"RHash\\rhashrc" };
+	static ctpath_t home_conf[3] = { L"$HOMEDRIVE", L"$HOMEPATH", L"rhashrc" };
+
+	if (try_config(app_data, COUNTOF(app_data)) || try_config(home_conf, COUNTOF(home_conf))) {
+		return 1;
+	} else {
+		tpath_t prog_dir[2];
+		prog_dir[0] = get_program_dir();
+		prog_dir[1] = L"rhashrc";
+		return try_config((ctpath_t*)prog_dir, COUNTOF(prog_dir));
 	}
 
-	/* then check for %HOMEDRIVE%%HOMEPATH%\rhashrc */
-	/* note that %USERPROFILE% is generally not a user home dir */
-	if ( (dir1 = getenv("HOMEDRIVE")) && (path = getenv("HOMEPATH"))) {
-		dir1 = make_path(dir1, path, 0);
-		path = make_path(dir1, CONFIG_FILENAME, 0);
-		free(dir1);
-		if (is_regular_file(path)) {
-			rsh_vector_add_ptr(opt.mem, path);
-			return (conf_opt.config_file = path);
-		}
-		free(path);
-	}
-
-	/* check for ${PROGRAM_DIR}\rhashrc */
-	if (program_dir && program_dir[0] && (dir1 = convert_wcs_to_str(program_dir, ConvertToPrimaryEncoding))) {
-		path = make_path(dir1, CONFIG_FILENAME, 0);
-		free(dir1);
-		if (is_regular_file(path)) {
-			rsh_vector_add_ptr(opt.mem, path);
-			return (conf_opt.config_file = path);
-		}
-		free(path);
-	}
 #endif /* _WIN32 */
-
-	return (conf_opt.config_file = NULL); /* config file not found */
 }
 
 /**
@@ -643,20 +683,19 @@ static int read_config(void)
 {
 #define LINE_BUF_SIZE 2048
 	char buf[LINE_BUF_SIZE];
-	file_t file;
 	FILE* fd;
 	parsed_option_t option;
 	int res;
 
-	/* initialize conf_opt and opt structures */
+	/* initialize conf_opt */
 	memset(&conf_opt, 0, sizeof(opt));
 	conf_opt.find_max_depth = -1;
 
 	if (!find_conf_file()) return 0;
+	assert(!!rhash_data.config_file.real_path);
+	assert(FILE_ISREG(&rhash_data.config_file));
 
-	file_init_by_print_path(&file, 0, conf_opt.config_file, 0);
-	fd = file_fopen(&file, FOpenRead);
-	file_cleanup(&file);
+	fd = file_fopen(&rhash_data.config_file, FOpenRead);
 	if (!fd) return -1;
 
 	while (fgets(buf, LINE_BUF_SIZE, fd)) {
@@ -671,7 +710,7 @@ static int read_config(void)
 		/* search for '=' */
 		index = strcspn(line, "=");
 		if (line[index] == 0) {
-			log_warning(_("%s: can't parse line \"%s\"\n"), conf_opt.config_file, line);
+			log_warning(_("%s: can't parse line \"%s\"\n"), file_get_print_path(&rhash_data.config_file, FPathUtf8 | FPathNotNull), line);
 			continue;
 		}
 		line[index] = 0;
@@ -684,7 +723,7 @@ static int read_config(void)
 		}
 
 		if (!t->type) {
-			log_warning(_("%s: unknown option \"%s\"\n"), conf_opt.config_file, line);
+			log_warning(_("%s: unknown option \"%s\"\n"), file_get_print_path(&rhash_data.config_file, FPathUtf8 | FPathNotNull), line);
 			continue;
 		}
 
@@ -709,7 +748,8 @@ static int read_config(void)
 	res = fclose(fd);
 
 #ifdef _WIN32
-	if ( (opt.flags & OPT_ENCODING) == 0 ) opt.flags |= (conf_opt.flags & OPT_ENCODING);
+	if ( (opt.flags & OPT_ENCODING) == 0 )
+		opt.flags |= (conf_opt.flags & OPT_ENCODING);
 #endif
 	return (res == 0 ? 0 : -1);
 }
@@ -1007,11 +1047,13 @@ static void set_default_sums_flags(const char* progName)
 	else if (strstr(buf, "gost94"))  res |= RHASH_GOST94;
 	if (strstr(buf, "has160"))  res |= RHASH_HAS160;
 	if (strstr(buf, "ripemd160") || strstr(buf, "rmd160"))  res |= RHASH_RIPEMD160;
-	if (strstr(buf, "whirlpool"))  res |= RHASH_WHIRLPOOL;
-	if (strstr(buf, "edonr256"))   res |= RHASH_EDONR256;
-	if (strstr(buf, "edonr512"))   res |= RHASH_EDONR512;
-	if (strstr(buf, "snefru256"))  res |= RHASH_SNEFRU128;
-	if (strstr(buf, "snefru128"))  res |= RHASH_SNEFRU256;
+	if (strstr(buf, "whirlpool")) res |= RHASH_WHIRLPOOL;
+	if (strstr(buf, "edonr256"))  res |= RHASH_EDONR256;
+	if (strstr(buf, "edonr512"))  res |= RHASH_EDONR512;
+	if (strstr(buf, "blake2s"))   res |= RHASH_BLAKE2S;
+	if (strstr(buf, "blake2b"))   res |= RHASH_BLAKE2B;
+	if (strstr(buf, "snefru256")) res |= RHASH_SNEFRU128;
+	if (strstr(buf, "snefru128")) res |= RHASH_SNEFRU256;
 	if (strstr(buf, "ed2k-link")) res |= OPT_ED2K_LINK;
 	else if (strstr(buf, "ed2k")) res |= RHASH_ED2K;
 
@@ -1072,9 +1114,9 @@ static void make_final_options_checks(void)
 {
 	unsigned ext_format_bits = (opt.printf_str ? 0x100 : 0) | (opt.template_file ? 0x200 : 0);
 
-	if ((opt.flags & OPT_VERBOSE) && conf_opt.config_file) {
+	if ((opt.flags & OPT_VERBOSE) && !!rhash_data.config_file.real_path) {
 		/* note that the first log_msg call shall be made after setup_output() */
-		log_msg(_("Config file: %s\n"), conf_opt.config_file);
+		log_msg_file_t(_("Config file: %s\n"), &rhash_data.config_file);
 	}
 
 	if (opt.bt_batch_file)
